import importlib

import pytest

from fed_rag import generators, knowledge_stores, retrievers


@pytest.mark.parametrize("name", generators.__all__)
def test_public_generators_all_importable(name: str) -> None:
    """Tests that all names listed in generators __all__ are importable."""
    mod = importlib.import_module("fed_rag.generators")
    attr = getattr(mod, name)

    assert hasattr(mod, name)
    assert attr is not None


@pytest.mark.parametrize("name", retrievers.__all__)
def test_public_retrievers_all_importable(name: str) -> None:
    """Tests that all names listed in retrievers __all__ are importable."""
    mod = importlib.import_module("fed_rag.retrievers")
    attr = getattr(mod, name)

    assert hasattr(mod, name)
    assert attr is not None


@pytest.mark.parametrize("name", knowledge_stores.__all__)
def test_public_knowledge_stores_all_importable(name: str) -> None:
    """Tests that all names listed in knowledge_stores __all__ are importable."""
    mod = importlib.import_module("fed_rag.knowledge_stores")
    attr = getattr(mod, name)

    assert hasattr(mod, name)
    assert attr is not None



from typing import Any

import pytest
import torch
from pydantic import PrivateAttr

from fed_rag import RAGConfig, RAGSystem
from fed_rag.base.generator import BaseGenerator
from fed_rag.base.retriever import BaseRetriever
from fed_rag.base.tokenizer import BaseTokenizer
from fed_rag.knowledge_stores.in_memory import InMemoryKnowledgeStore


class MockRetriever(BaseRetriever):
    _encoder: torch.nn.Module = PrivateAttr(default=torch.nn.Linear(3, 3))

    def encode_context(self, context: str, **kwargs: Any) -> torch.Tensor:
        return self._encoder.forward(torch.ones(3))

    def encode_query(self, query: str, **kwargs: Any) -> torch.Tensor:
        return self._encoder.forward(torch.zeros(3))

    @property
    def encoder(self) -> torch.nn.Module:
        return self._encoder

    @property
    def query_encoder(self) -> torch.nn.Module | None:
        return None

    @property
    def context_encoder(self) -> torch.nn.Module | None:
        return None


@pytest.fixture
def mock_retriever() -> MockRetriever:
    return MockRetriever()


class MockTokenizer(BaseTokenizer):
    def encode(self, input: str, **kwargs: Any) -> list[int]:
        return [0, 1, 2]

    def decode(self, input_ids: list[int], **kwargs: Any) -> str:
        return "mock decoded sentence"

    @property
    def unwrapped(self) -> None:
        return None


@pytest.fixture()
def mock_tokenizer() -> BaseTokenizer:
    return MockTokenizer()


class MockGenerator(BaseGenerator):
    _model = torch.nn.Linear(2, 1)
    _tokenizer = MockTokenizer()
    _prompt_template = "{query} and {context}"

    def generate(self, query: str, context: str, **kwargs: Any) -> str:
        return f"mock output from '{query}' and '{context}'."

    def compute_target_sequence_proba(self, prompt: str, target: str) -> float:
        return 0.42

    @property
    def model(self) -> torch.nn.Module:
        return self._model

    @model.setter
    def model(self, value: torch.nn.Module) -> None:
        self._model = value

    @property
    def tokenizer(self) -> MockTokenizer:
        return self._tokenizer

    @tokenizer.setter
    def tokenizer(self, value: BaseTokenizer) -> None:
        self._tokenizer = value

    @property
    def prompt_template(self) -> str:
        return self._prompt_template

    @prompt_template.setter
    def prompt_template(self, v: str) -> None:
        self._prompt_template = v


@pytest.fixture
def mock_generator() -> BaseGenerator:
    return MockGenerator()


@pytest.fixture
def mock_rag_system(
    mock_generator: BaseGenerator,
    mock_retriever: BaseRetriever,
) -> RAGSystem:
    knowledge_store = InMemoryKnowledgeStore()
    rag_config = RAGConfig(
        top_k=2,
    )
    return RAGSystem(
        generator=mock_generator,
        retriever=mock_retriever,
        knowledge_store=knowledge_store,
        rag_config=rag_config,
    )



import importlib

import pytest

from fed_rag import __all__ as _root_all


@pytest.mark.parametrize("name", _root_all)
def test_root_names_all_importable(name: str) -> None:
    """Tests that all names listed in root __all__ are importable."""
    mod = importlib.import_module("fed_rag")
    attr = getattr(mod, name)

    assert hasattr(mod, name)
    assert attr is not None


def test_no_internal_leakage() -> None:
    """Test that internal implementation details aren't exposed in the API."""
    # Check that no private/internal classes are exposed
    for name in _root_all:
        assert not name.startswith("_"), f"API exposes internal name: {name}"

    # no base classes in api
    assert (
        "BaseBridgeMixin" not in _root_all
    ), "API shouldn't expose base classes"
    assert (
        "BaseDataCollator" not in _root_all
    ), "API shouldn't expose base classes"
    assert "BaseFLTask" not in _root_all, "API shouldn't expose base classes"
    assert (
        "BaseGenerator" not in _root_all
    ), "API shouldn't expose base classes"
    assert (
        "BaseKnowledgeStore" not in _root_all
    ), "API shouldn't expose base classes"
    assert (
        "BaseRetriever" not in _root_all
    ), "API shouldn't expose base classes"
    assert (
        "BaseTokenizer" not in _root_all
    ), "API shouldn't expose base classes"
    assert (
        "BaseTrainerConfig" not in _root_all
    ), "API shouldn't expose base classes"
    assert "BaseTrainer" not in _root_all, "API shouldn't expose base classes"
    assert (
        "BaseGeneratorTrainer" not in _root_all
    ), "API shouldn't expose base classes"
    assert (
        "BaseRetrieverGenerator" not in _root_all
    ), "API shouldn't expose base classes"
    assert (
        "BaseRAGTrainerManager" not in _root_all
    ), "API shouldn't expose base classes"



"""Public Generators API"""

# Disable the F403 warning for wildcard imports
# ruff: noqa: F403, F401
from .huggingface import *
from .huggingface import __all__ as _huggingface_all

__all__ = _huggingface_all



import importlib
from contextlib import nullcontext as does_not_raise

import pytest

from fed_rag import types


def test_root_imports() -> None:
    """Test that core types can be imported from the root."""
    with does_not_raise():
        # ruff: noqa: F401
        from fed_rag import RAGConfig, RAGSystem


def test_type_imports() -> None:
    """Test that there are no circular imports in the types module."""
    with does_not_raise():
        # ruff: noqa: F401
        from fed_rag.types import (  # RAGConfig,; RAGSystem,
            KnowledgeNode,
            NodeContent,
            NodeType,
            RAGConfig,
            RAGResponse,
            SourceNode,
            TestResult,
            TrainResult,
        )


def test_base_direct_imports() -> None:
    """Test that base classes can be imported directly."""
    with does_not_raise():
        # ruff: noqa: F401
        from fed_rag.base.bridge import BaseBridgeMixin
        from fed_rag.base.data_collator import BaseDataCollator
        from fed_rag.base.fl_task import BaseFLTask
        from fed_rag.base.generator import BaseGenerator
        from fed_rag.base.knowledge_store import BaseKnowledgeStore
        from fed_rag.base.retriever import BaseRetriever
        from fed_rag.base.tokenizer import BaseTokenizer
        from fed_rag.base.trainer import (
            BaseGeneratorTrainer,
            BaseRetrieverTrainer,
            BaseTrainer,
        )
        from fed_rag.base.trainer_config import BaseTrainerConfig
        from fed_rag.base.trainer_manager import BaseRAGTrainerManager


def test_all_simultaneously() -> None:
    """Test that there are no circular imports."""
    with does_not_raise():
        # ruff: noqa: F401
        from fed_rag import RAGConfig, RAGSystem
        from fed_rag.base.bridge import BaseBridgeMixin
        from fed_rag.base.data_collator import BaseDataCollator
        from fed_rag.base.fl_task import BaseFLTask
        from fed_rag.base.generator import BaseGenerator
        from fed_rag.base.knowledge_store import BaseKnowledgeStore
        from fed_rag.base.retriever import BaseRetriever
        from fed_rag.base.tokenizer import BaseTokenizer
        from fed_rag.base.trainer import (
            BaseGeneratorTrainer,
            BaseRetrieverTrainer,
            BaseTrainer,
        )
        from fed_rag.base.trainer_config import BaseTrainerConfig
        from fed_rag.base.trainer_manager import BaseRAGTrainerManager
        from fed_rag.types import (
            BridgeMetadata,
            KnowledgeNode,
            NodeContent,
            NodeType,
            RAGResponse,
            SourceNode,
            TestResult,
            TrainResult,
        )


@pytest.mark.parametrize("name", types.__all__)
def test_types_all_importable(name: str) -> None:
    """Tests that all names listed in __all__ are importable."""
    mod = importlib.import_module("fed_rag.types")
    attr = getattr(mod, name)

    assert hasattr(mod, name)
    assert attr is not None


def test_import_rag_system_from_types() -> None:
    """Test deprecated import path for RAGSystem from fed_rag.types"""

    with pytest.warns(DeprecationWarning):
        # ruff: noqa: F401
        from fed_rag.types.rag_system import RAGSystem



"""The fed-rag library: simplified fine-tuning for RAG systems"""

from fed_rag._version import VERSION

# Disable the F403 warning for wildcard imports
# ruff: noqa: F403, F401
from .core import *
from .core import __all__ as _core_all
from .generators import *
from .generators import __all__ as _generators_all
from .knowledge_stores import *
from .knowledge_stores import __all__ as _knowledge_stores_all
from .retrievers import *
from .retrievers import __all__ as _retrievers_all
from .types import RAGConfig

__version__ = VERSION


__all__ = sorted(
    _core_all
    + _generators_all
    + _knowledge_stores_all
    + _retrievers_all
    + ["RAGConfig"]
)



"""Auxiliaty types for RAG System"""

from typing import Any

from pydantic import BaseModel

from .knowledge_node import KnowledgeNode


class SourceNode(BaseModel):
    score: float
    node: KnowledgeNode

    def __getattr__(self, __name: str) -> Any:
        """Convenient wrapper on getattr of associated node."""
        return getattr(self.node, __name)


class RAGResponse(BaseModel):
    response: str
    source_nodes: list[SourceNode]

    def __str__(self) -> str:
        return self.response


class RAGConfig(BaseModel):
    top_k: int
    context_separator: str = "\n"



"""Internal RAG System Module"""

from typing import TYPE_CHECKING, Any, ClassVar

from pydantic import BaseModel, ConfigDict

from fed_rag.base.bridge import BridgeMetadata
from fed_rag.types import KnowledgeNode, RAGConfig, RAGResponse, SourceNode

if TYPE_CHECKING:  # pragma: no cover
    # to avoid circular imports, using forward refs
    from fed_rag.base.generator import BaseGenerator
    from fed_rag.base.knowledge_store import BaseKnowledgeStore
    from fed_rag.base.retriever import BaseRetriever


class _RAGSystem(BaseModel):
    """Unbridged implementation of RAGSystem.

    IMPORTANT: This is an internal implementation class.
    It should only be used by bridge mixins and never referenced directly
    by user code or other parts of the library.

    All interaction with RAG systems should be through the public RAGSystem class.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    generator: "BaseGenerator"
    retriever: "BaseRetriever"
    knowledge_store: "BaseKnowledgeStore"
    rag_config: RAGConfig
    bridges: ClassVar[dict[str, BridgeMetadata]] = {}

    @classmethod
    def _register_bridge(cls, metadata: BridgeMetadata) -> None:
        """To be used only by `BaseBridgeMixin`."""
        if metadata["framework"] not in cls.bridges:
            cls.bridges[metadata["framework"]] = metadata

    def query(self, query: str) -> RAGResponse:
        """Query the RAG system."""
        source_nodes = self.retrieve(query)
        context = self._format_context(source_nodes)
        response = self.generate(query=query, context=context)
        return RAGResponse(source_nodes=source_nodes, response=response)

    def retrieve(self, query: str) -> list[SourceNode]:
        """Retrieve from KnowledgeStore."""
        query_emb: list[float] = self.retriever.encode_query(query).tolist()
        raw_retrieval_result = self.knowledge_store.retrieve(
            query_emb=query_emb, top_k=self.rag_config.top_k
        )
        return [
            SourceNode(score=el[0], node=el[1]) for el in raw_retrieval_result
        ]

    def generate(self, query: str, context: str) -> str:
        """Generate response to query with context."""
        return self.generator.generate(query=query, context=context)  # type: ignore

    def _format_context(self, source_nodes: list[SourceNode]) -> str:
        """Format the context from the source nodes."""
        # TODO: how to format image context
        return str(
            self.rag_config.context_separator.join(
                [node.get_content()["text_content"] for node in source_nodes]
            )
        )


def _resolve_forward_refs() -> None:
    """Resolve forward references in _RAGSystem."""

    # These imports are needed for Pydantic to resolve forward references
    # ruff: noqa: F401
    from fed_rag.base.generator import BaseGenerator
    from fed_rag.base.knowledge_store import BaseKnowledgeStore
    from fed_rag.base.retriever import BaseRetriever

    # Update forward references
    _RAGSystem.model_rebuild()


_resolve_forward_refs()



"""Bridge type definitions for fed-rag."""

from typing import TypedDict


class BridgeMetadata(TypedDict):
    """Type definition for bridge metadata."""

    bridge_version: str
    framework: str
    compatible_versions: list[str]
    method_name: str



"""RAG System Module"""

from fed_rag._bridges.llamaindex.bridge import LlamaIndexBridgeMixin
from fed_rag.core._rag_system import _RAGSystem


# Define the public RAGSystem with all available bridges
class RAGSystem(LlamaIndexBridgeMixin, _RAGSystem):
    """RAG System with all available bridge functionality.

    The RAGSystem is the main entry point for creating and managing
    retrieval-augmented generation systems.
    """

    pass
